// OpenALTestDlg.cpp : implementation file
//

#include "stdafx.h"
#include "OpenALTest.h"
#include "OpenALTestDlg.h"

#include "sndfile.h"
#include "al.h"
#include "alc.h"
#include "Delay.h"
#include "Downsampler.h"
#include "Filter.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif


// COpenALTestDlg dialog


COpenALTestDlg::COpenALTestDlg(CWnd* pParent /*=NULL*/)
	: CDialog(COpenALTestDlg::IDD, pParent)
	, fileName(_T("Enter a .wav file name here (mono PCM)"))
	, slider1(0)
	, slider2(0)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	stopping = false;
	numThreads = 0;
	device = alcOpenDevice(NULL); // TODO: error handling
	context = alcCreateContext(device, NULL);
	alcMakeContextCurrent(context);
	fileNameMutex = CreateMutex(NULL, FALSE, NULL);
	sliderMutex = CreateMutex(NULL, FALSE, NULL);
}

void COpenALTestDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);

	WaitForSingleObject(fileNameMutex, INFINITE);
	DDX_Text(pDX, IDC_FILE_NAME, fileName);
	DDV_MaxChars(pDX, fileName, 128);
	ReleaseMutex(fileNameMutex);
	DDX_Control(pDX, IDC_SLIDER1, cSlider1);
	DDX_Control(pDX, IDC_SLIDER2, cSlider2);
}

BEGIN_MESSAGE_MAP(COpenALTestDlg, CDialog)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
	ON_BN_CLICKED(IDC_START, &COpenALTestDlg::OnBnClickedStart)
	ON_BN_CLICKED(IDC_STOP, &COpenALTestDlg::OnBnClickedStop)
	ON_WM_VSCROLL()
END_MESSAGE_MAP()


// COpenALTestDlg message handlers

BOOL COpenALTestDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	cSlider1.SetRange(0, 100);
    cSlider1.SetPos(100);
	cSlider1.SetTicFreq(10);

	cSlider2.SetRange(0, 100);
    cSlider2.SetPos(100);
	cSlider2.SetTicFreq(10);

	return TRUE;  // return TRUE  unless you set the focus to a control
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void COpenALTestDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR COpenALTestDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}

COpenALTestDlg::~COpenALTestDlg()
{
	stopping = true;

	// InterlockedCompareExchange only used to retrieve the value
	while(0 != InterlockedCompareExchange(&numThreads, 0, 0))
	{
		Sleep(50);
	}

	alcMakeContextCurrent(NULL);
	alcDestroyContext(context);
	alcCloseDevice(device);

	CloseHandle(sliderMutex);
	CloseHandle(fileNameMutex);
}

DWORD WINAPI processWaveFile(LPVOID lpParameter)
{
	COpenALTestDlg* parent = (COpenALTestDlg*)lpParameter;

#ifdef _DEBUG
	{
		char s[128];
		sprintf(s, "Thread %d started\n", GetCurrentThreadId()),
		OutputDebugStringA(s);
	}
#endif

	SF_INFO sfinfo;
	sfinfo.format = SF_FORMAT_WAV;
	
	WaitForSingleObject(parent->fileNameMutex, INFINITE);
	int sizeOfString = (parent->fileName.GetLength() + 1);
	LPTSTR lpsz = new TCHAR[sizeOfString];
	_tcsncpy(lpsz, parent->fileName, sizeOfString);
	ReleaseMutex(parent->fileNameMutex);

	SNDFILE* sndFile = sf_open(lpsz, SFM_READ, &sfinfo);
	delete[] lpsz;
	if(sndFile == NULL || sfinfo.channels != 1)
	{
#ifdef _DEBUG
		{
			char s[128];
			sprintf(s, "Thread %d aborted immediately\n", GetCurrentThreadId()),
			OutputDebugStringA(s);
		}
#endif
		InterlockedDecrement(&parent->numThreads);
		return -1;
	}

	const int NUMBUFFERS = 4;
	ALuint bufferIDs[NUMBUFFERS];
	alGenBuffers(NUMBUFFERS, bufferIDs);
	ALuint sourceID;
	alGenSources(1, &sourceID);

	const ALuint BSIZE = 1024;
	short* dataIn = new short[BSIZE];
	short* dataOut = new short[BSIZE];
	ALuint dataSize = (ALuint) sfinfo.frames;

	Delay myFilter(BSIZE);

	bool starting = true;

	while(!parent->stopping)
	{
		if (dataSize == 0) // no more samples left in the file
		{
			break;
		}

		ALint numBuffersToRefill;
		ALuint buffersToRefill[NUMBUFFERS];
		if(starting) // on startup, fill _all_ buffers, even though they're not yet processed
		{
			numBuffersToRefill = NUMBUFFERS;
			for(int b = 0; b < NUMBUFFERS; b++)
			{
				buffersToRefill[b] = bufferIDs[b];
			}
			starting = false;
		}
		else
		{
			alGetSourcei(sourceID, AL_BUFFERS_PROCESSED, &numBuffersToRefill);
			alSourceUnqueueBuffers(sourceID, numBuffersToRefill, buffersToRefill);
		}

		if(numBuffersToRefill == 0)
		{
			// don't burn too much CPU power on waiting for free buffers;
			// some sort of notification wuold be better (if it existed)
			Sleep(10);
		}
		else
		{
#ifdef _DEBUG
			{
				char s[128];
				char s1[128];
				sprintf(s, "Thread %d fills buffer(s):", GetCurrentThreadId());
				for(int b = 0; b < numBuffersToRefill; b++)
				{
					strcpy(s1, s);
					sprintf(s, "%s %d", s1, buffersToRefill[b]);
				}
				strcpy(s1, s);
				sprintf(s, "%s\n", s1),
				OutputDebugStringA(s);
			}
#endif

			// get fresh data
			for(int b = 0; b < numBuffersToRefill; b++)
			{
				ALuint dataToRead = (dataSize > BSIZE) ? BSIZE : dataSize;
				if(dataToRead > 0)
				{
					dataSize -= dataToRead;
					sf_read_short(sndFile, dataIn, dataToRead);
					myFilter.Process(dataIn, dataOut, dataToRead, parent);
				}
				alBufferData(buffersToRefill[b], AL_FORMAT_MONO16, dataOut, dataToRead*sizeof(short), sfinfo.samplerate);
			}
			alSourceQueueBuffers(sourceID, numBuffersToRefill, buffersToRefill);
		}

		ALint playing;
		alGetSourcei(sourceID, AL_SOURCE_STATE, &playing);
		if (playing != AL_PLAYING)
		{	
			alSourcePlay(sourceID);
		}
	}

	alSourceStop(sourceID);
	alDeleteSources(1, &sourceID);

	alDeleteBuffers(NUMBUFFERS, bufferIDs);

	sf_close(sndFile);

	delete[] dataIn;
	delete[] dataOut;

	// now the OpenAL device may safely be removed
	InterlockedDecrement(&parent->numThreads);

#ifdef _DEBUG
	{
		char s[128];
		sprintf(s, "Thread %d terminated\n", GetCurrentThreadId()),
		OutputDebugStringA(s);
	}
#endif

	return 0;
}

void COpenALTestDlg::OnBnClickedStart()
{
	UpdateData(TRUE);
	HANDLE workerThread = CreateThread(NULL, 0, processWaveFile, this, 0, NULL);
	CloseHandle(workerThread);
	InterlockedIncrement(&numThreads);
}

void COpenALTestDlg::OnBnClickedStop()
{
	stopping = true;

	// InterlockedCompareExchange only used to retrieve the value
	while(0 != InterlockedCompareExchange(&numThreads, 0, 0))
	{
		Sleep(50);
	}

	stopping = false;
}

void COpenALTestDlg::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
	if(nSBCode == SB_THUMBPOSITION ||nSBCode == SB_THUMBTRACK)
	// We can't handle the other nSBCodes so easily: They don't have nPos data.
	// So always drag the slider, don't click and don't use the keyboard shortcuts.   :-(
	{
		switch(pScrollBar->GetDlgCtrlID())
		{
		case IDC_SLIDER1:
			WaitForSingleObject(sliderMutex, INFINITE);
			slider1 = 1.0 - 0.01*nPos;
			ReleaseMutex(sliderMutex);
			break;
		case IDC_SLIDER2:
			WaitForSingleObject(sliderMutex, INFINITE);
			slider2 = 1.0 - 0.01*nPos;
			ReleaseMutex(sliderMutex);
			break;
		}
	}
	CDialog::OnVScroll(nSBCode, nPos, pScrollBar);
}
